
import re
import threading
import time

from smithy.chain import InvocationChain, EmptyInvocationChain
from smithy.filelist import FileList

class Task(object):
    def __init__(self, app, name):
        self.app = app
        self.mgr = app.mgr
        self.scope = app.mgr.scope[:]
        self.name = name
        self.descr = None
        self.sources = []
        self.predeps = []
        self.actions = []
        self.already_invoked = False
        self.lock = threading.RLock()

    def __str__(self):
        return self.name
    
    def __repr__(self):
        klass = self.__class__.__name__
        return "<%s %s => [%s]>" % (klass, self.name, ', '.join(self.deps))

    @classmethod
    def scoped_name(klass, scope, task_name):
        return ":".join(scope + [task_name])
    
    @property
    def source(self):
        "The first source or None if no sources exist."
        if len(self.sources):
            return self.sources[0]
        return None
    
    @property
    def deps(self):
        expanded = []
        for d in self.predeps:
            if isinstance(d, FileList):
                expanded.extend(d.expand())
            else:
                expanded.append(d)
        return expanded

    @property
    def as_source(self):
        """\
        Return the source representation of this task. Mostly meant for
        FileTasks to return the filename they are expected to produce.
        """
        return None

    def clear(self):
        "Clear the existing prerequisites and actions of this task."
        while len(self.actions):
            self.actions.pop()
        while len(self.deps):
            self.deps.pop()
        return self

    def enhance(self, action=None, deps=None, description=None):
        "Add an action and/or dependencies to this task."
        if action is not None:
            self.actions.append(action)
        if deps is not None:
            self.predeps.extend(deps)
        if description is not None:
            # Describing a task always overrides the current
            # description.
            self.descr = description
        elif self.descr is None:
            # Only use a doc string if we don't already have
            # a description.
            if hasattr(action, "func_doc") and action.func_doc:
                self.descr = action.func_doc
        return self

    def execute(self, args=None):
        "Execute the actions associated with this task."
        args = args or tuple() #TaskArgs.EMPTY
        if self.app.is_dry_run():
            self.app.log("** Execute (dry run) %s" % self.name)
            return
        self.app.log("** Execute %s" % self.name)
        if not len(self.actions):
            self.mgr.add_from_rule(self.name)
        self.sources = self._get_sources()
        for act in self.actions:
            if act.func_code.co_argcount == 0:
                ret = act()
            elif act.func_code.co_argcount == 1:
                ret = act(self)
            else:
                ret = act(self, *args)
            self.app.log_output(self.name, ret)

    def invoke(self, *args):
        "Invoke the task if it is needed. Prerequisites are invoked first."
        #args = TaskArguments(self.arg_names, args)
        self._invoke(args, EmptyInvocationChain())

    def needed(self):
        "Is this task needed?"
        return True
    
    def reenable(self):
        "Allow this task to be invoked again even if it has already run."
        self.already_invoked = False
    
    def _invoke(self, args, chain):
        chain = InvocationChain(chain, self)
        with self.lock:
            if self.already_invoked:
                return
            self.already_invoked = True
            self._invoke_deps(args, chain)
            if self.needed():
                self.execute(args)
    
    def _invoke_deps(self, args, chain):
        for dep in self.deps:
            prereq = self.mgr.find(dep, self.scope)
            #prereq_args = args.new_scope(prereq.arg_names)
            prereq._invoke(args, chain)

    def _get_sources(self):
        tasks = map(lambda d: self.mgr.find(d, self.scope), self.deps)
        tasks = filter(lambda t: t.as_source, tasks)
        return map(lambda t: t.name, tasks)                

    def _trace_info(self):
        "Format trace flags for display."
        flags = []
        if not self.already_invoked:
            flags.append("first_time")
        if not self.needed():
            flags.append("not_needed")
        if len(flags):
            return "(%s)" % ', '.join(flags)
        return ""
    
    def dump(self):
        """\
        Return a string describing the internal state of this task. Useful
        for debugging.
        """
        result = ["-" * 40]
        result.append("Investigation %s" % self.name)
        result.append("class: %s" % self.__class__.__name__)
        result.append("task needed: %s" % self.needed())
        result.append("timestamp: %s" % self.timestamp())
        result.append("pre-requisites: ")
        prereqs = map(lambda p: self.mgr.lookup(p), self._prereqs)
        prereqs.sort(cmp=lambda a, b: cmp(a.timestamp(), b.timestamp()))
        for req in prereqs:
            result.append("--%s (%s)" % req.name, req.timestamp())
        result.append("." * 40)
        result.extend(["", ""])
        return '\n'.join(result)



# Copyright 2010 Paul J. Davis <paul.joseph.davis@gmail.com>
#
# This file is part of the Smithy package released under the MIT license.

import t

@t.test
def test_trigger_single_dep():
    @t.task
    def foo():
        return 2
    @t.task([foo])
    def bar():
        return "hi"
    t.app.run(None, ["bar"])
    t.eq(t.app._dbg, [("foo", 2), ("bar", "hi")])

@t.test
def test_doesnt_trigger_children():
    @t.task
    def foo():
        return 2
    @t.task(["foo"])
    def bar():
        return "hi"
    t.app.run(None, ["foo"])
    t.eq(t.app._dbg, [("foo", 2)])

@t.test
def test_not_rerun_as_dep():
    @t.task
    def foo():
        return 2
    @t.task(["foo"])
    def bar():
        return "hi"
    t.app.run(None, ["foo"])
    t.eq(t.app._dbg, [("foo", 2)])
    t.app.run(None, ["bar"])
    t.eq(t.app._dbg, [("foo", 2), ("bar", "hi")])

@t.test
def test_multi_deps():
    @t.task
    def foo():
        return 2
    @t.task
    def bar():
        return "hi"
    @t.task(["foo", bar])
    def frog():
        return "bird"
    t.app.run(None, ["frog"])
    t.eq(t.app._dbg, [
        ("foo", 2),
        ("bar", "hi"),
        ("frog", "bird")
    ])

@t.test
def test_diamond_deps():
    @t.task
    def foo():
        return 2
    @t.task([foo])
    def bar():
        return "hi"
    @t.task(["foo"])
    def frog():
        return "bird"
    @t.task([bar, frog])
    def proton():
        return False
    t.app.run(None, ["proton"])
    t.eq(t.app._dbg, [
        ("foo", 2),
        ("bar", "hi"),
        ("frog", "bird"),
        ("proton", False)
    ])




# Copyright 2010 Paul J. Davis <paul.joseph.davis@gmail.com>
#
# This file is part of the Smithy package released under the MIT license.

import t

@t.test
def test_basic_description():
    @t.task
    def foo():
        "bar"
    t.eq(t.app.mgr.lookup("foo").descr, "bar")

@t.test
def test_add_description():
    @t.task
    def foo():
        pass
    t.desc("foo", "baz!")
    t.eq(t.app.mgr.lookup("foo").descr, "baz!")

@t.test
def test_add_description_first():
    t.desc("foo", "bazinga")
    @t.task
    def foo():
        "Not used as a description"
        pass
    t.eq(t.app.mgr.lookup("foo").descr, "bazinga")

@t.test
def test_override_docstring():
    @t.task
    def foo():
        "Not the description"
        pass
    t.desc("foo", "wheeee")
    t.eq(t.app.mgr.lookup("foo").descr, "wheeee")

@t.test
def test_from_func():
    @t.task
    def foo():
        "Hi"
        pass
    t.desc(foo, "yay")
    t.eq(t.app.mgr.lookup("foo").descr, "yay")




class SmithyError(Exception):
    pass

class NoBellowsFileError(SmithyError):
    def __str__(self):
        return "Failed to find a Bellows file."

class TaskArgumentError(SmithyError):
    "Illformed task declaration"
    pass

class TaskNotFoundError(KeyError, SmithyError):
    def __init__(self, name):
        self.name = name
    def __str__(self):
        return "Task not found: %s" % self.name

class NoActionForTaskError(RuntimeError, SmithyError):
    "No action can be found or created for a given task."
    pass

class DependencyCycleError(RuntimeError, SmithyError):
    "Cyclic invocation chain."
    def __init__(self, name):
        self.name = name
    def __str__(self):
        return "Dependency cycle detected at: %s" % self.name

class NoRuleError(SmithyError):
    "Unable to recursively generate a task for this source."
    pass

class RuleRecursionOverflowError(SmithyError):
    "Recursion overflow in task selection"
    def __init__(self, *args):
        super(RuleRecursionOverflowError, self).__init__(*args)
        self.targets = []

    def __str__(self):
        base = super(RuleRecursionOverflowError, self).__str__()
        return "%s : [%s]" % (base, ' => '.join(reversed(self.targets)))

    def add_target(self, target):
        self.targets.append(target)





import copy
import functools
import glob
import os
import re

from smithy.path import aspath, path

def _exclude_cores(fn):
    return re.search(r"(^|[\/\\])core$", fn) and not os.path.isdir(fn)

class FileListIter(object):
    def __init__(self, source):
        self.source = source
    def next(self):
        return aspath(self.source.next())

class FileList(object):
    """\
    A FileList is essentially an special class for dealing with lists of
    file names.
    
    FileLists are lazy. When given a list of glob patterns for possible
    files to be included in the file list, instead of search the file
    structures to find the files, a FileList holds the pattern for latter use.
    
    This allows us to define a number of patterns to match any number of
    files, but only search out the actual files when the FileList itself
    is actually used.
    """
    
    DEFAULT_EXCLUDE_PATTERNS = [
        re.compile(r"(^|[\/\\])CVS([\/\\]|$)"),
        re.compile(r"(^|[\/\\])\.svn([\/\\]|$)"),
        re.compile(r"(^|[\/\\])\.git([\/\\]|$)"),
        re.compile(r"(^|[\/\\])\.hg([\/\\]|$)"),
        re.compile(r"\.bak$"),
        re.compile(r"~$")
    ]
    
    DEFAULT_EXCLUDE_FUNCS = [
        _exclude_cores
    ]
    
    def __init__(self, *patterns):
        """\
        Create a FileList from the provided globbable patterns. If you
        wish to perform multiple includes or excludes when creating
        new instances, use the method chaining pattern.
        
        Example:
          file_list = FileList("lib/**/*.py", "test/test*.py")
          
          pkg_files = FileList("lib/**/*").exclude_re(r"\bCVS\b")
                        .exclude_re(r"\b.git\b").exclude_re("\b.svn\b")
        """
        if len(patterns) == 1 and isinstance(patterns[0], FileList):
            other = patterns[0]
            self._items = copy.copy(other._items)
            self._pending_add = copy.copy(other._pending_add)
            self._pending = other._pending
            self._exclude_patterns = copy.copy(other._exclude_patterns)
            self._exclude_funcs = copy.copy(other._exclude_funcs)
            self._exclude_regexps = copy.copy(other._exclude_regexps)
        else:
            self._items = []
            self._pending_add = []
            self._pending = False
            self._exclude_patterns = FileList.DEFAULT_EXCLUDE_PATTERNS[:]
            self._exclude_funcs = FileList.DEFAULT_EXCLUDE_FUNCS[:]
            self._exclude_regexps = []
            map(self.include, patterns)
    
    # Not hashable
    
    __hash__ = None
    
    # Printing functions
    
    def __str__(self):
        self.resolve()
        return '[%s]' % ', '.join(map(str, self))
    
    def __repr__(self):
        return "<FileList %s>" % str(self)
    
    # Binary operator methods
    
    def __add__(self, other):
        print "Adding: %r" % other
        if not isinstance(other, (list, FileList)):
            return NotImplemented
        elif isinstance(other, FileList):
            self.resolve()
            other.resolve()
            self._items += map(aspath, other._items)
        else:
            self.resolve()
            self._items += map(aspath, other)
        return self
    
    def __radd__(self, other):
        if not isinstance(other, (list, FileList)):
            return NotImplemented
        elif isinstance(other, FileList):
            self.resolve()
            other.resolve
            self._items = other._items + self._items
        else:
            self.resolve()
            self._items = map(aspath, other) + self._items
        return self

    # Container Methods

    def __len__(self):
        self.resolve()
        return len(self._items)

    def __iter__(self):
        self.resolve()
        return FileListIter(iter(self._items))

    def __getitem__(self, item):
        self.resolve()
        if not isinstance(item, slice):
            return aspath(self._items[item])
        ret = FileList(self)
        ret._items = self._items[item]
        return ret

    def __setitem__(self, item, value):
        self.resolve()
        if not isinstance(item, slice):
            self._items[item] = aspath(value)
        elif isinstance(value, list):
            self._items[item] = map(aspath, value)
        elif isinstance(value, FileList):
            value.resolve()
            self._items[item] = value._items
        else:
            raise TypeError("%r is not an instance of list or FileList")

    def __delitem__(self, item):
        self.resolve()
        del self._items[item]

    def __contains__(self, item):
        self.resolve()
        return aspath(item) in self._items
    
    # Comparison functions
    
    def __lt__(self, other):
        if isinstance(other, FileList):
            self.resolve()
            other.resolve()
            return self._items < other._items
        elif isinstance(other, list):
            self.resolve()
            return self._items < other
        return NotImplemented
    
    def __le__(self, other):
        if isinstance(other, FileList):
            self.resolve()
            other.resolve()
            return self._items <= other._items
        elif isinstance(other, list):
            self.resolve()
            return self._items <= other
        return NotImplemented
    
    def __eq__(self, other):
        if isinstance(other, FileList): 
            self.resolve()
            other.resolve()
            return self._items == other._items
        elif isinstance(other, list):
            self.resolve()
            return self._items == other
        return NotImplemented
    
    def __ne__(self, other):
        if isinstance(other, FileList):
            self.resolve()
            other.resolve()
            return self._items != other._items
        elif isinstance(other, list):
            self.resolve()
            return self._items != other
        return NotImplemented
    
    def __ge__(self, other):
        if isinstance(other, FileList):
            self.resolve()
            other.resolve()
            return self._items >= other._items
        elif isinstance(other, list):
            self.resolve()
            return self._items >= other
        return NotImplemented
    
    def __gt__(self, other):
        if isinstance(other, FileList):
            self.resolve()
            other.resolve()
            return self._items >= other._items
        elif isinstance(other, list):
            self.resolve()
            return self._items >= other
        return NotImplemented

    # List interface functions.

    def append(self, obj):
        self.resolve()
        self._items.append(aspath(obj))
    
    def count(self, obj):
        self.resolve()
        return self._items.count(aspath(obj))
    
    def extend(self, iterobj):
        self.resolve()
        self._items.extend(aspath(p) for p in iterobj)
    
    def index(self, obj, *args):
        self.resolve()
        return self._items.index(aspath(obj), *args)

    def insert(self, index, obj):
        self.resolve()
        self._items.insert(index, aspath(obj))

    def pop(self, index):
        self.resolve()
        return aspath(self._items.pop(index))

    def remove(self, obj):
        self.resolve()
        self._items.remove(aspath(obj))

    def reverse(self):
        self.resolve()
        self._items.reverse()
    
    def sort(self, *args, **kwargs):
        self.resolve()
        self._items.sort(*args, **kwargs)

    # FileList extension functions.

    def copy(self):
        "Return a cloned instance of this FileList"
        return self.__class__(self)

    def expand(self):
        ret = self.__class__(self)
        ret.resolve()
        return ret

    def include(self, *patterns):
        """\
        Add new glob patterns to this FileList instance. If an aray is given,
        add each element of the array.
        
        Example:
          file_list.include("*.rb", "*.cfg")
          file_list.include("math.c lib.h *.o".split())
        """
        for p in patterns:
            if isinstance(p, basestring):
                self._pending_add.append(p)
            else:
                map(self.include, p)
        self._pending = True
        return self

    def exclude(self, *patterns):
        """\
        Register a list of glob patterns that should be excluded from the
        list. Remember that a full pathname is a valid glob if you want
        to exclude specific files.
        
        Note that glob patterns are expanded against the file system. If a
        file is explicitly aded to af ile list, but does not exist on the
        file system, then a glob pattern in the exclude list will not
        exclude the file.
        
        Examples:
          FileList('a.c', 'b.c').exclude('a.c') -> ['b.c']
         
        If "a.c" is a file:
          FileList('a.c', 'b.c').exclude('a.*') -> ['b.c']
         
        If "a.c" is not a file:
          FileList('a.c', 'b.c').exclude('a.*') -> ['a.c', 'b.c']
        
        """
        for p in patterns:
            if isinstance(p, basestring):
                self._exclude_patterns.append(p)
            elif callable(p):
                self._exclude_funcs.append(p)
            else:
                map(self.exclude, p)
        if not self._pending:
            self._resolve_excludes()
        return self
    
    def exclude_re(self, *patterns):
        """\
        Register a list of regular expression patterns or objects that
        should be excluded from the list of file names.
        
        The `search` method is used to compare regular expressions
        to file names which means that you need to explicitly anchor
        patterns with '^' if you want to start at the beginning of
        the string.
        
        Examples:
          FileList('ab.c', 'b.c').exclude_re(r"a.*") -> ['b.c']
          FileList('ab.c', 'b.c').exclude_re(r"^b") -> ['ab.c']
        
        """
        for p in patterns:
            if isinstance(p, basestring):
                self._exclude_patterns.append(re.compile(p))
            elif hasattr(p, "pattern") and hasattr(p, "search"):
                # Can't find a better test for a compiled regular expression
                self._exclude_patterns.append(p)
            else:
                map(self.exclude_re, p)
        if not self._pending:
            self._resolve_excludes()
        return self
    
    def is_excluded(self, fn):
        "Should the given filename be excluded?"
        if not self._exclude_regexps:
            self._calculate_exclude_regexp()
        if any(r.search(fn) for r in self._exclude_regexps):
            return True
        return any(func(fn) for func in self._exclude_funcs)
    
    def clear_excludes(self):
        "Clear all the exclude patterns."
        self._exclude_patterns = []
        self._exclude_funcs = []
        if not self._pending:
            self._calculate_exclude_regexp()
        return self

    def resolve(self):
        "Resolve all the pending glob patterns."
        if self._pending:
            self._pending = False
            for fn in self._pending_add:
                self._resolve_add(fn)
            self._pending_add = []
            self._resolve_excludes()
        return self

    def sub(self, pattern, repl):
        """\
        Calls re.sub(pattern, repl, fn) on each filename.

        Triggers file name resolution.
        """
        self.resolve()
        p = re.compile(pattern)
        for idx in range(len(self)):
            self[idx] = p.sub(repl, self[idx])
        return self

    def exist(self):
        """\
        Remove any file names that don't exist on the file system.
        
        Triggers file name resolution.
        """
        self.resolve()
        idx = 0
        while idx < len(self):
            if not self[idx].exists():
                self.pop(idx)
            else:
                idx += 1
        return self

    def _resolve_add(self, fn):
        if re.search(r"[*?\[\{]", fn):
            self._add_matching(fn)
        else:
            self.append(aspath(fn))

    def _add_matching(self, pattern):
        "Add files matching the glob pattern."
        fnames = filter(lambda fn: not self.is_excluded(fn), glob.glob(pattern))
        self.extend(aspath(fn) for fn in fnames)

    def _exclude(self, fn):
        "Should the given filename be excluded?"
        if not self._exclude_regexps:
            self._calculate_exclude_regexp()
        if any(r.search(fn) for r in self._exclude_regexps):
            return True
        return any(func(fn) for func in self._exclude_funcs)
    
    def _resolve_excludes(self):
        self._calculate_exclude_regexp()
        idx = 0
        while idx < len(self):
            if self.is_excluded(self[idx]):
                self.pop(idx)
            else:
                idx += 1

    def _calculate_exclude_regexp(self):
        self._exclude_regexps = []
        for e in self._exclude_patterns:
            if hasattr(e, "pattern") and hasattr(e, "search"):
                self._exclude_regexps.append(e)
            elif re.search(r"[*?]", e):
                for fn in glob.glob(e):
                    self._exclude_regexps.append(re.compile(fn))
            else:
                self._exclude_regexps.append(re.compile(e))



# Copyright 2010 Paul J. Davis <paul.joseph.davis@gmail.com>
#
# This file is part of the Smithy package released under the MIT license.

import t

@t.test
def test_no_tasks():
    t.raises(t.TaskNotFoundError, t.app.run, None, ["foo"])

@t.test
def test_one_task():
    @t.task
    def foo():
        return 3
    t.app.run(None, ["foo"])
    t.eq(t.app._dbg, [("foo", 3)])

@t.test
def test_two_tasks():
    @t.task
    def foo():
        return 2
    @t.task
    def bar():
        return "hi"
    t.app.run(None, ["foo", "bar"])
    t.eq(t.app._dbg, [("foo", 2), ("bar", "hi")])

@t.test
def test_task_arg():
    @t.task
    def bizzle(task):
        from smithy.task import Task
        t.eq(isinstance(task, Task), True)
        return task.name
    t.app.run(None, ["bizzle"])
    t.eq(t.app._dbg, [("bizzle", "bizzle")])

@t.test
def test_task_run_once():
    @t.task
    def foo():
        return 2
    t.app.run(None, ["foo"])
    t.eq(t.app._dbg, [("foo", 2)])
    t.app.run(None, ["foo"])
    t.eq(t.app._dbg, [("foo", 2)])

@t.test
def test_no_decorator():
    def foo():
        return 2
    t.task("foo")(foo)
    t.app.run(None, ["foo"])
    t.eq(t.app._dbg, [("foo", 2)])

@t.test
def test_adding_more_actions():
    def foo():
        return 2
    def bar():
        return "hi"
    t.task("foo")(foo)
    t.task("foo")(bar)
    t.app.run(None, ["foo"])
    t.eq(t.app._dbg, [("foo", 2), ("foo", "hi")])
    t.raises(t.TaskNotFoundError, t.app.run, None, ["bar"])

@t.test
def test_lambda_action():
    print t.task(lambda: 4.5)
    print t.app.mgr.tasks
    t.app.run(None, ["<lambda>.1"])
    t.eq(t.app._dbg, [("<lambda>.1", 4.5)])



# Copyright 2010 Paul J. Davis <paul.joseph.davis@gmail.com>
#
# This file is part of the Smithy package released under the MIT license.

import t

@t.test
def test_basic_rule():
    csrc = t.tmpfname("foo.c")
    cobj = t.tmpfname("foo.o")
    app = t.tmpfname("foo")
    @t.rule(".o", ".c")
    def compile(task):
        open(task.name, "wb").write(open(task.source).read() + "+with_o")
    @t.build(app, [cobj])
    def link(task):
        open(task.name, "wb").write(open(task.source).read() + "+linked")
    open(csrc, "wb").write("source")
    t.app.run(None, [app])
    t.eq(t.app._dbg, [(cobj, None), (app, None)])

@t.test
def test_multi_sources():
    csrcs = [t.tmpfname() + ".c" for i in range(10)]
    cobjs = [fn[:-2] + ".o" for fn in csrcs]
    app = t.tmpfname()
    @t.rule(".o", ".c")
    def compile(task):
        open(task.name, "wb").write(open(task.source).read() + ".o")
    @t.build(app, cobjs)
    def link(task):
        open(task.name, "wb").write(open(task.source).read() + ".exe")
    for cs in csrcs:
        open(cs, "wb").write("source")
    t.app.run(None, [app])
    t.eq(len(t.app._dbg), len(cobjs)+1)
    t.eq(t.app._dbg[0], (cobjs[0], None))
    t.eq(t.app._dbg[-1], (app, None))

def make_chain(depth):
    if depth < 1:
        fname = t.tmpfname()
        open(fname + ".0", "wb").write("ohai!")
        return fname
    @t.rule(".%d" % depth, ".%d" % (depth-1))
    def chained(task):
        open(task.name, "wb").write(open(task.source).read())
    return make_chain(depth-1)

@t.test
def test_chained_rules():
    fname = make_chain(14)
    @t.build(fname + ".15", [fname+".14"])
    def cap(task):
        open(task.name, "wb").write("capped!")
    t.app.run(None, [fname+".15"])
    t.eq(len(t.app._dbg), 15) # .0 is not recorded as run
    t.eq(open(fname+".15").read(), "capped!")


# Copyright 2010 Paul J. Davis <paul.joseph.davis@gmail.com>
#
# This file is part of the Smithy package released under the MIT license.

import t

@t.test
def test_simple_cycle():
    @t.task(["foo"])
    def foo():
        return 2
    t.raises(t.DependencyCycleError, t.app.run, None, ["foo"])

@t.test
def test_two_task_cycle():
    t.task("foo", ["bar"])(lambda: 2)
    t.task("bar", ["foo"])(lambda: "hi")
    
    for task in ["foo", "bar"]:
        t.app.clear()
        t.raises(t.DependencyCycleError, t.app.run, None, [task])
        t.eq(t.app._dbg, [])
    
@t.test
def test_ring_cycle():
    t.task("foo")(lambda: 2)
    t.task("bar")(lambda: "hi")
    t.task("frog")(lambda: "bird")
    
    t.task("foo", ["frog"])
    t.task("bar", ["foo"])
    t.task("frog", ["bar"])

    for task in ["foo", "bar", "frog"]:
        t.app.clear()
        t.raises(t.DependencyCycleError, t.app.run, None, [task])
        t.eq(t.app._dbg, [])


# Copyright 2010 Paul J. Davis <paul.joseph.davis@gmail.com>
#
# This file is part of the Smithy package released under the MIT license.

import t

@t.filetest
def test_create_empty():
    t.eq(len(t.FileList()), 0)

@t.filetest
def test_create_with_args():
    fl = t.FileList("*.c", "x")
    t.eq(sorted("abc.c x.c xyz.c x".split()), sorted(fl))

@t.filetest
def test_str():
    fl = t.FileList("a.c", "b.c", "x.py")
    t.eq(str(fl), "[a.c, b.c, x.py]")
    t.eq("%s" % fl, "[a.c, b.c, x.py]")

@t.filetest
def test_str_resolve():
    fl = t.FileList("*.c")
    str(fl)
    t.eq(len(fl), 3)

@t.filetest
def test_delayed_resolve():
    # Check sanity
    fl = t.FileList("*.c")
    t.eq(len(fl), 3)
    
    # Check basic delay
    fl = t.FileList("*.c")
    t.eq(len(fl._items), 0)
    str(fl)
    t.eq(len(fl._items), 3)
    
    # Check trigger by second reference
    fl = t.FileList("*.c")
    t.eq(len(fl._items), 0)
    fl2 = fl
    t.eq(len(fl._items), 0)
    t.eq(len(fl._items), 0)
    repr(fl2)
    t.eq(len(fl._items), 3)
    t.eq(len(fl._items), 3)
    
    # Check after copy
    fl = t.FileList("*.c")
    t.eq(len(fl._items), 0)
    fl2 = fl.copy()
    t.eq(len(fl._items), 0)
    t.eq(len(fl2._items), 0)
    repr(fl2)
    t.eq(len(fl._items), 0)
    t.eq(len(fl2._items), 3)

    # Check expand doesn't affect the original
    fl3 = fl.expand()
    t.eq(len(fl._items), 0)
    t.eq(len(fl3._items), 3)
    t.eq(fl2 == fl3, True)

@t.filetest
def test_reresolve():
    fl = t.FileList("a.c", "*.c")
    for i in range(10):
        fl.resolve()
        t.eq(len(fl), 4)

@t.filetest
def test_create_from_iterable():
    fl = t.FileList(["x", "y", "z"])
    t.eq(sorted(fl), sorted(["x", "y", "z"]))

@t.filetest
def test_include_other_file_list():
    fl = t.FileList(t.FileList("*.c", "x"))
    t.eq(sorted(fl), sorted(["abc.c", "x.c", "xyz.c", "x"]))

@t.filetest
def test_append():
    fl = t.FileList()
    fl.append("a.py")
    fl.append("b.py")
    t.eq(fl, ["a.py", "b.py"])

@t.filetest
def test_add_many():
    fl = t.FileList()
    fl.include("a x c".split())
    fl.include("d", "y")
    t.eq(sorted(fl), "a c d x y".split())

@t.filetest
def test_match():
    fl = t.FileList()
    fl.include("*.c")
    t.isin("abc.c", fl)
    t.eq(len(fl), 3)
    map(lambda p: t.eq(p.ext, ".c"), fl)

@t.filetest
def test_add_matching():
    fl = t.FileList()
    fl.append("a.java")
    fl.include("*.c")
    t.eq(fl[0], "a.java")
    t.eq(len(fl), 4)
    t.isin("abc.c", fl)

@t.filetest
def test_multiple_patterns():
    fl = t.FileList("foo/*.c", "bar/*xist*")
    t.eq(fl, [])
    fl.include("*.c", "*xist*")
    t.eq(sorted(fl), sorted("abc.c existing x.c xyz.c".split()))

@t.filetest
def test_square_bracket_pattern():
    fl = t.FileList("abc.[ch]")
    t.eq(sorted(fl), ["abc.c", "abc.h"])

@t.filetest
def test_filter():
    fl = t.FileList()
    fl.include("x.c abc.c xyz.c existing".split())
    fl = filter(lambda p: "x" not in p, fl)
    t.eq(sorted(fl), ["abc.c"])

@t.filetest
def test_exclude():
    fl = t.FileList("x.c", "abc.c", "xyz.c", "existing")
    x = fl.exclude_re(r"x.+\.")
    t.isinstance(x, t.FileList)
    t.eq(sorted(fl), ["abc.c", "existing", "x.c"])
    t.eq(id(fl), id(x))
    fl.exclude("*.c")
    t.eq(fl, ["existing"])
    fl.exclude("existing")
    t.eq(fl, [])

@t.filetest
def test_exclude_via_callable():
    fl = t.FileList("a.c", "b.c", "xyz.c")
    fl.exclude(lambda x: x == "xyz.c")
    t.eq(fl, ["a.c", "b.c"])

@t.filetest
def test_exclude_re():
    fl = t.FileList("*").exclude_re(r".*\.[hcx]$")
    t.eq(sorted(fl), ["existing", "x"])
    t.isinstance(fl, t.FileList)

@t.filetest
def test_exclude_filename():
    fl = t.FileList("*.c").exclude("abc.c")
    t.eq(sorted(fl), ["x.c", "xyz.c"])
    t.isinstance(fl, t.FileList)

@t.filetest
def test_default_excludes():
    fl = t.FileList()
    fl.clear_excludes()
    fl.include("*~", "*.bak", "core")
    t.isin("core", fl)
    t.isin("x.bak", fl)
    t.isin("x~", fl)

@t.filetest
def test_clear_excludes():
    fl = t.FileList("*")
    t.isnotin("CVS", fl)
    t.isnotin(".svn", fl)
    t.isnotin(".dummy", fl)
    t.isnotin("x.bak", fl)
    t.isnotin("x~", fl)
    t.isnotin("core", fl)    

@t.filetest
def test_exclude_alternate_file_seps():
    fl = t.FileList()
    t.eq(fl.is_excluded("x/CVS/y"), True)
    t.eq(fl.is_excluded("x\\CVS\\y"), True)
    t.eq(fl.is_excluded("x/.svn/y"), True)
    t.eq(fl.is_excluded("x\\.svn\\y"), True)
    t.eq(fl.is_excluded("x/.git/y"), True)
    t.eq(fl.is_excluded("x\\.git\\y"), True)
    t.eq(fl.is_excluded("x/core"), True)
    t.eq(fl.is_excluded("x\\core"), True)

    # Core as path component is kosher.
    t.eq(fl.is_excluded("x/core/y"), False)
    t.eq(fl.is_excluded("x\\core\\y"), False)

@t.filetest
def test_add_to_exclude():
    fl = t.FileList()
    fl.exclude_re(r"~\d+")
    t.eq(fl.is_excluded("x/CVS/y"), True)
    t.eq(fl.is_excluded("x\\.git\\y"), True)
    t.eq(fl.is_excluded("x/core"), True)
    t.eq(fl.is_excluded("abc~3"), True)

@t.filetest
def test_sub():
    fl = t.FileList("*.c")
    fl.sub(r".c$", ".o")
    t.eq(sorted(fl), ["abc.o", "x.o", "xyz.o"])

@t.filetest
def test_sub_with_backref():
    fl = t.FileList("src/org/onestepback/a.java", "src/org/onestepback/b.java")
    fl.sub(r"^src/(.*).java$", r"classes/\1.class")
    t.eq(sorted(fl), [
        "classes/org/onestepback/a.class",
        "classes/org/onestepback/b.class"
    ])

@t.filetest
def test_exist():
    fl = t.FileList("abc.c", "notthere.c")
    t.eq(fl.exist(), ["abc.c"])

@t.filetest
def test_doesnt_subclass_list():
    fl = t.FileList("*.c")
    t.eq(isinstance(fl, list), False)

@t.filetest
def test_not_hashable():
    fl = t.FileList("*.c")
    t.raises(TypeError, hash, fl)

@t.filetest
def test_add_arrays():
    t.eq(t.FileList("a.c", "b.c") + ["foo.py"], ["a.c", "b.c", "foo.py"])
    t.eq(["foo.py"] + t.FileList("a.c", "b.c"), ["foo.py", "a.c", "b.c"])
    t.raises(TypeError, lambda: t.FileList("a.c", "b.c") + 1)
    t.raises(TypeError, lambda: t.FileList("a.c", "b.c") + [1])
    t.raises(TypeError, lambda: 1 + t.FileList("a.c", "b.c"))
    t.raises(TypeError, lambda: [1] + t.FileList("a.c", "b.c"))

@t.filetest
def test_length():
    t.eq(len(t.FileList("a.c", "b.c")), 2)
    t.eq(len(t.FileList("*.c")), 3)

@t.filetest
def test_iter():
    fl = t.FileList("a.c", "b.c", "exist*")
    i = iter(fl)
    t.eq(isinstance(i, t.FileListIter), True)
    t.eq(i.next(), "a.c")
    t.eq(isinstance(i.next(), t.path), True)
    t.eq(i.next(), "existing")
    t.raises(StopIteration, i.next)

@t.filetest
def test_get_item():
    fl = t.FileList("a.c", "b.c")
    t.eq(fl[0], "a.c")
    t.eq(fl[-1], "b.c")
    t.raises(IndexError, lambda: fl[100])

@t.filetest
def test_set_item():
    fl = t.FileList("a.c", "b.c")
    fl[0] = "foo.py"
    t.eq(fl, ["foo.py", "b.c"])
    def assign():
        fl[1] = 2
    t.raises(TypeError, assign)

@t.filetest
def test_del_item():
    fl = t.FileList("a.c", "b.c")
    del fl[0]
    t.eq(fl[0], "b.c")
    def rem():
        del fl[2]
    t.raises(IndexError, rem)

@t.filetest
def test_contains():
    fl = t.FileList("*.c")
    t.isin("abc.c", fl)
    t.isin("x.c", fl)
    t.isin("xyz.c", fl)

@t.filetest
def test_get_slice():
    fl = t.FileList("a.c", "b.c", "exist*")
    fl1 = fl[1:]
    t.eq(fl1, ["b.c", "existing"])
    t.eq(isinstance(fl1, t.FileList), True)

@t.filetest
def test_set_slice():
    fl1 = t.FileList("a.c", "b.c", "exist*")
    fl2 = t.FileList("d.c")
    fl1[1:2] = fl2
    t.eq(fl1, ["a.c", "d.c", "existing"])
    fl1[2:] = ["d.e"]
    t.eq(fl1, ["a.c", "d.c", "d.e"])

@t.filetest
def test_del_slice():
    fl = t.FileList("a.c", "b.c", "exist*")
    del fl[1:]
    t.eq(fl, ["a.c"])

@t.filetest
def test_comparisons():
    fl1 = t.FileList("a.c", "exist*")
    fl2 = t.FileList("b.c")
    fl3 = t.FileList("a.c", "exist*")

    # __lt__
    t.eq(fl1 < fl2, True)
    t.eq(fl2 < fl1, False)
    t.eq(fl1 < ["d.e"], True)
    t.eq(["a.c"] < fl2, True)
    t.eq(["z.x"] < fl1, False)

    # __le__
    t.eq(fl1 <= fl1, True)
    t.eq(fl1 <= fl2, True)
    t.eq(fl2 <= fl1, False)
    t.eq(fl1 <= ["d.e"], True)
    t.eq(["_"] <= fl1, True)
    t.eq(["z.c"] <= fl2, False)
    
    # __eq__
    t.eq(fl1 == fl3, True)
    t.eq(fl3 == fl1, True)
    t.eq(fl2 == fl2, True)
    t.eq(fl2 == ["b.c"], True)
    t.eq(fl1 == fl2, False)
    t.eq(fl1 == ["z"], False)
    t.eq(["b.c"] == fl2, True)
    t.eq(["z"] == fl1, False)
    
    # __ne__
    t.eq(fl1 != fl2, True)
    t.eq(fl1 != fl1, False)
    t.eq(fl1 != fl3, False)
    t.eq(fl1 != ["a.c", "existing"], False)
    t.eq(fl2 != fl3, True)
    t.eq(["z.b"] != fl2, True)
    t.eq(["a.c", "existing"] != fl1, False)
    
    # __ge__
    t.eq(fl1 >= fl3, True)
    t.eq(fl2 >= fl1, True)
    t.eq(fl1 >= fl2, False)
    t.eq(fl1 >= ["_"], True)
    t.eq(fl2 >= ["z.c"], False)
    t.eq(["z.c"] >= fl2, True)
    t.eq(["b.c"] >= fl2, True)
    t.eq(["a"] >= fl1, False)
    
    # __gt__
    t.eq(fl2 > fl1, True)
    t.eq(fl1 > fl2, False)
    t.eq(fl2 > ["a.c"], True)
    t.eq(fl2 > ["z.c"], False)
    t.eq(["z.c"] > fl1, True)
    t.eq(["a"] > fl3, False)



# Copyright 2010 Paul J. Davis <paul.joseph.davis@gmail.com>
#
# This file is part of the Smithy package released under the MIT license.

import t

@t.test
def test_simple_ns():
    with t.ns("rho"):
        @t.task
        def foo():
            return 2
    t.raises(t.TaskNotFoundError, t.app.run, None, ["foo"])
    t.app.run(None, ["rho:foo"])
    t.eq(t.app._dbg, [("rho:foo", 2)])

def check_ns():
    t.raises(t.TaskNotFoundError, t.app.run, None, ["foo"])
    t.raises(t.TaskNotFoundError, t.app.run, None, ["rho:foo"])
    t.app.run(None, ["rho:wade:foo"])
    t.eq(t.app._dbg, [("rho:wade:foo", 2)])

@t.test
def test_nested_ns():
    with t.ns("rho"):
        with t.ns("wade"):
            @t.task
            def foo():
                return 2
    check_ns()

@t.test
def test_nested_ns_one_with():
    with t.ns("rho", "wade"):
        @t.task
        def foo():
            return 2 
    check_ns()

@t.test
def test_nested_reference():
    with t.ns("rho"):
        @t.task
        def foo():
            return 2
        with t.ns("wade"):
            @t.task([foo])
            def bar():
                return "hi"
    t.app.run(None, ["rho:wade:bar"])
    



import types

from smithy.application import application
from smithy.task import Task
from smithy.filelist import FileList
from smithy.filetask import FileTask, FileCreationTask
from smithy.multitask import MultiTask
from smithy.path import aspath

__all__ = ["desc", "rule", "task", "build", "multitask", "ns"]

FUNC_TYPES = (types.FunctionType, types.BuiltinFunctionType)
METH_TYPES = (types.MethodType, types.BuiltinMethodType)
LAMBDA_COUNT = 0
def _get_name(func):
    global LAMBDA_COUNT
    if isinstance(func, basestring):
        return func
    elif isinstance(func, METH_TYPES):
        return func.im_func.func_name
    elif isinstance(func, FUNC_TYPES):
        if func.func_name == "<lambda>":
            LAMBDA_COUNT += 1
            return "<lambda>.%d" % LAMBDA_COUNT
        return func.func_name
    elif isinstance(func, FileList):
        return func
    else:
        raise TypeError("Unable to make a task from type: %s" % func.__class__)

def desc(name, description):
    application.mgr.describe(_get_name(name), description)

def rule(pattern, sources):
    def _create_rule(func):
        application.mgr.add_rule(pattern, sources, func)
        return func
    return _create_rule

def task(*args, **kwargs):
    name = kwargs.get("name", None)
    type = kwargs.get("type", Task)
    action = kwargs.get("action", None)
    deps = kwargs.get("deps", None)
    
    if len(args) == 1 and isinstance(args[0], (list, tuple)) and deps is None:
        deps = map(_get_name, args[0])
    elif len(args) == 1:
        if action is None and callable(args[0]):
            action = args[0]
        if name is None:
            name = _get_name(args[0])
    elif len(args) == 2:
        if action is None and callable(args[0]):
            action = args[0]
        if name is None:
            name = _get_name(args[0])
        if deps is None and isinstance(args[1], (list, tuple)):
            deps = map(_get_name, args[1])
    elif len(args) != 0:
        raise TaskArgumentError()

    if type is not None and name is not None:
        t = application.mgr.lookup(name)
        if t is None:
            application.mgr.add_task(type, name, deps=deps)
        elif deps is not None:
            t.enhance(deps=deps)
        # Don't re-add deps.
        deps = None
    
    if action is None: # Need to recurse
        return lambda f: task(f, type=type, name=name, deps=deps)
    
    application.mgr.add_task(type, name, action=action, deps=deps)
    return action

def build(fname, *args, **kwargs):
    type = FileTask if kwargs.get("recreate", True) else FileCreationTask
    return task(*args, type=type, name=aspath(fname))

def multitask(func):
    return task(func, type=MultiTask)

class ns(object):
    def __init__(self, *args):
        self.scopes = map(str, args)
    
    def __enter__(self):
        map(application.mgr.push_scope, self.scopes)
        return self

    def __exit__(self, type, value, traceback):
        map(application.mgr.pop_scope, reversed(self.scopes))



# Copyright 2010 Paul J. Davis <paul.joseph.davis@gmail.com>
#
# This file is part of the Smithy package released under the MIT license.

import os
import t

@t.test
def test_not_rebuilt():
    prefile = t.tmpfname("foo.c")
    postfile = t.tmpfname("bar.c")
    src = t.tmpfname()
    tgt = t.tmpfname()
    
    # Make the prefile, if this is the only file
    # available to the mktgt rule, then we know that
    # lazy evaluation was broken
    with open(prefile, "wb") as out:
        out.write("prefile")
    
    @t.build(src)
    def mksrc(task, recreate=True):
        task.name.write_bytes(t.uuid.uuid4().hex)
        postfile.write_bytes("postfile")
    @t.build(tgt, [src, t.FileList(t.DATA_DIR / "*.c")])
    def mktgt(task):
        return sorted(task.sources)
        t.eq(sorted(task.sources), )
        out.write_bytes(data.bytes())

    # Check that postfile doesn't exist yet.
    t.eq(postfile.exists(), False)

    # Run the pipeline, test that target saw the
    # expected sources
    t.app.run(None, [tgt])
    tgt_out = sorted([src, t.DATA_DIR/"foo.c", t.DATA_DIR/"bar.c"])
    t.eq(t.app._dbg, [(src, None), (tgt, tgt_out)])
    
    # Postfile should now exist:
    t.eq(postfile.exists(), True)

    # Rerun without destroying postfile
    t.app.clear()
    t.app.run(None, [tgt])
    tgt_out = sorted([src, t.DATA_DIR/"foo.c", t.DATA_DIR/"bar.c"])
    t.eq(t.app._dbg, [(src, None), (tgt, tgt_out)])
    
    # Create a new file to show that we rebind
    # for each run.
    
    t.tmpfname("zoo.c").write_bytes("postpostfile")
    
    t.app.clear()
    t.app.run(None, [tgt])
    tgt_out = sorted([src, t.DATA_DIR/"foo.c",
                        t.DATA_DIR/"bar.c", t.DATA_DIR/"zoo.c"])
    t.eq(t.app._dbg, [(src, None), (tgt, tgt_out)])
    



from smithy.exceptions import DependencyCycleError

class InvocationChain(object):    
    def __init__(self, chain, task):
        self.parent = chain
        self.known = chain.known.copy()
        if task in self.known:
            raise DependencyCycleError(task)
        self.known.add(task)
        self.task = task
    
    def __str__(self):
        return ', '.join(self.task_names())

    def task_names(self):
        if not self.parent:
            return []
        return self.parent.task_names() + [self.task.name]

class EmptyInvocationChain(InvocationChain):
    def __init__(self):
        self.parent = None
        self.known = set()
        self.task = None




import os
import re

from smithy.exceptions import \
        TaskNotFoundError, NoActionForTaskError, NoRuleError
from smithy.filetask import FileCreationTask, FileTask
from smithy.task import Task

class TaskManager(object):
    
    def __init__(self, app):
        self.app = app
        self.tasks = {}
        self.rules = []
        self.scope = []
        self.last_description = None
    
    def add_from_rule(self, name, depth=0):
        if depth > self.app.rule_depth():
            raise RuleRecursionOverflowError(name)
        for (pattern, extensions, action) in self.rules:
            if not pattern.search(name):
                continue
            return self._apply_rule(name, extensions, action, depth+1)
        return None

    def add_rule(self, pattern, deps, action):
        if isinstance(pattern, basestring):
            pattern = re.compile(re.escape(pattern) + "$")
        if not isinstance(deps, (list, tuple)):
            deps = [deps]
        self.rules.append((pattern, deps, action))
    
    def add_task(self, type, name, action=None, deps=None):
        name = type.scoped_name(self.scope, name)
        task = self._intern(type, name)
        task.enhance(action=action, deps=deps)
        return task

    def describe(self, name, description):
        task = self.lookup(name)
        if task is None:
            task = self.add_task(Task, name)
        task.enhance(description=description)

    def find(self, name, scope=None):
        task = self.lookup(name, scope)
        if task is not None:
            return task
        task = self.add_from_rule(name)
        if task is not None:
            return task
        task = self._make_file_task(name)
        if task is not None:
            return task
        raise NoActionForTaskError(name)

    def lookup(self, name, scope=None):
        curr_scope = scope or self.scope
        curr_scope = curr_scope[:] # Don't change the original
        if name.startswith("trowel:"):
            name = name[len("trowel:"):]
            curr_scope = []
        elif name.startswith("^"):
            while name.startswith("^"):
                if len(curr_scope): curr_scope.pop(-1)
                name = name[1:]
        return self._lookup(name, curr_scope)

    def push_scope(self, scope):
        self.scope.append(scope)
    
    def pop_scope(self, scope):
        if not len(self.scope):
            raise IndexError("Pop from empty scope list.")
        if scope != self.scope[-1]:
            raise ValueError("%s != %s" % self.scope[-1], scope)
        self.scope.pop(-1)
    
    def _intern(self, type, name):
        if name not in self.tasks:
            self.tasks[name] = type(self.app, name)
        return self.tasks[name]
    
    def _lookup(self, name, scopes):
        while len(scopes):
            tn = ':'.join(scopes + [name])
            task = self.tasks.get(tn)
            if task is not None:
                return task
            scopes.pop(-1)
        return self.tasks.get(name)

    def _make_file_task(self, task_name):
        if not os.path.isfile(task_name):
            return None
        return self.add_task(FileCreationTask, task_name)
    
    def _apply_rule(self, name, extensions, action, depth):
        def _mk_task(src):
            self.app.trace("** Applying %s => %s" % (name, src))
            if os.path.exists(src):
                self.app.trace("** Exists %s => %s" % (name, src))
                return src
            parent = self.add_from_rule(src, depth=depth)
            if parent is not None:
                return parent.name
            self.app.trace("** Failed %s => %s" % (name, src))
            raise NoRuleError(name, src)
        deps = map(_mk_task, self._make_sources(name, extensions))
        return self.add_task(FileTask, name, action=action, deps=deps)
    
    def _make_sources(self, name, extensions):
        print "Make sources: %s" % name
        ret = []
        for ext in extensions:
            if ext[:1] == ".":
                (base, ignore) = os.path.splitext(name)
                ret.append("%s%s" % (base, ext))
            elif callable(ext):
                if ext.func_code.co_argcount == 1:
                    ret.append(ext(name))
                else:
                    ret.append(ext())
            else:
                raise TypeError("Invalid replacement type: %r" % ext)
        return ret




import os

from smithy.const import EARLY
from smithy.filelist import FileList
from smithy.task import Task

class FileTask(Task):

    @classmethod
    def scoped_name(klass, scope, task_name):
        return task_name
    
    @property
    def as_source(self):
        return self.name
    
    def needed(self):
        if not os.path.exists(self.name):
            return True
        if len(self.deps) == 0:
            return True
        our_stamp = self.timestamp()
        for d in self.deps:
            task = self.mgr.find(d)
            if not isinstance(task, FileTask):
                return True
            elif task.timestamp() > our_stamp:
                return True
        return False

    def timestamp(self):
        if os.path.exists(self.name):
            return os.stat(self.name).st_mtime
        return EARLY
    
class FileCreationTask(FileTask):
    
    def needed(self):
        return not os.path.exists(self.name)
    
    def timestamp(self):
        if os.path.exists(self.name):
            return os.stat(self.name).st_mtime
        return EARLY




# This file is part of Belt released under the MIT license. 
# See the NOTICE for more information.

import os
from setuptools import setup, find_packages

README = os.path.join(os.path.dirname(__file__), 'README.rst')

setup(
    name = 'Smithy',
    version = '0.3.1',

    description = 'Python port of Ruby Rake',
    long_description = open(README).read(),
    author = 'Paul J. Davis',
    author_email = 'paul.joseph.davis@gmail.com',
    license = 'MIT',
    url = 'http://packages.python.org/Smithy/',

    classifiers = [
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Topic :: Software Development :: Build Tools',
        'Topic :: Software Development :: Libraries',
    ],
    
    packages = find_packages(),
    include_package_data = True,
    scripts = ['bin/smithy'],
    
    test_suite = 'nose.collector',
)




# Copyright 2010 Paul J. Davis <paul.joseph.davis@gmail.com>
#
# This file is part of the Smithy package released under the MIT license.

import functools
import os
import unittest
import uuid

from smithy.application import application as app
from smithy.decorators import *
from smithy.exceptions import *
from smithy.filelist import FileList, FileListIter
from smithy.path import aspath, path

DATA_DIR = path(__file__).parent / "data"

FILE_PATHS = map(path, """
    CVS .svn .git .dummy x.bak x x~ core x.c xyz.c abc.c abc.h abc.x existing
""".split())

def tmpfname(name=None):
    if not DATA_DIR.isdir():
        DATA_DIR.makedirs()
    if name is None:
        name = uuid.uuid4().hex.upper()
    return DATA_DIR / name

def test(func):
    @functools.wraps(func)
    def _f(*args, **kwargs):
        # Reset the internal state before each test.
        app.clear(tasks=True)
        func(*args, **kwargs)
        if DATA_DIR.isdir():
            DATA_DIR.rmtree()
    return _f

def filetest(func):
    @functools.wraps(func)
    def _f(*args, **kwargs):
        for p in FILE_PATHS:
            p = DATA_DIR / "lists" / p
            if not p.parent.isdir():
                p.parent.makedirs()
            p.touch()
        cwd = path.getcwd()
        try:
            os.chdir(DATA_DIR/"lists")
            func(*args, **kwargs)
        finally:
            os.chdir(cwd)
        if DATA_DIR.isdir():
            DATA_DIR.rmtree()
    return _f
            
def eq(a, b):
    assert a == b, "%r != %r" % (a, b)

def ne(a, b):
    assert a != b, "%r == %r" % (a, b)

def lt(a, b):
    assert a < b, "%r >= %r" % (a, b)

def gt(a, b):
    assert a > b, "%r <= %r" % (a, b)

def isin(a, b):
    assert a in b, "%r is not in %r" % (a, b)

def isnotin(a, b):
    assert a not in b, "%r is in %r" % (a, b)

def has(a, b):
    assert hasattr(a, b), "%r has no attribute %r" % (a, b)

def hasnot(a, b):
    assert not hasattr(a, b), "%r has an attribute %r" % (a, b)

def raises(exctype, func, *args, **kwargs):
    try:
        func(*args, **kwargs)
    except exctype, inst:
        return
    func_name = getattr(func, "func_name", "<builtin_function>")
    mesg = "Function %s did not raise %s" % (func_name, exctype.__name__)
    raise AssertionError(mesg)

def isinstance(obj, type):
    func = __builtins__["isinstance"]
    assert func(obj, type), "%r is not an instance of %r." % (obj, type)




import glob
import optparse as op
import os
import re
import sys

import smithy.exceptions as exc
from smithy.filelist import FileList
from smithy.path import aspath, path
from smithy.taskmanager import TaskManager

__usage__ = "%prog [OPTIONS] task1 [task2 ...]"

class Application(object):
    def __init__(self):
        self.mgr = TaskManager(self)
        self.opts = None
        self.globals = {}
        self.required_files = []
        self.args = []
        self._dbg = []

    def clear(self, tasks=False):
        if tasks: self.mgr = TaskManager(self)
        for t in self.mgr.tasks.itervalues():
            t.reenable()
        self._dbg = []

    def run(self, opts, tasks):
        self.opts = opts
        
        if self.opts:
            self.load()
            if self.opts.list:
                self.display_tasks(tasks)
                exit(0)

        if len(tasks) == 0:
            print "No task specified."
            exit(0)

        for t in tasks:
            try:
                task = self.mgr.find(t)
            except exc.NoActionForTaskError:
                raise exc.TaskNotFoundError(t)
            if task is None:
                raise exc.TaskNotFoundError(t)
            task.invoke()

    def load(self):
        # Add Python paths
        for dn in self.opts.incdirs:
            sys.path.insert(0, dn)
        # Load the main file.
        bellows = self.find_bellows()
        # Jump to the directory we're running from.
        self.workdir = os.path.dirname(bellows)
        if self.workdir and self.workdir != os.getcwd():
            os.chdir(self.workdir)
            sys.path.insert(0, self.workdir)
        self.load_file(bellows)
        # Load files that were required
        while len(self.required_files):
            required = self.required_files[:]
            self.required_files = []
            map(self.load_file, required)
        # Load system and library tasks
        if self.opts.incsys:
            for fn in glob.glob(self.opts.sysglob):
                self.load_file(fn)
        for fn in glob.glob(self.opts.libglob):
            self.load_file(fn)

    def load_file(self, fname):
        self.globals[fname] = self.init_globals(fname)
        with open(fname) as handle:
            src = handle.read()
        code = compile(src, fname, 'exec')
        exec code in self.globals[fname]
    
    def init_globals(self, fname):
        import smithy.decorators as dec
        return {
            "__file__": fname,
            "aspath": aspath,
            "path": path,
            "FileList": FileList,
            "require": self.require,
            "desc": dec.desc,
            "task": dec.task,
            "rule": dec.rule,
            "build": dec.build,
            "multitask": dec.multitask,
            "ns": dec.ns
        }

    def find_bellows(self):
        BELLOWS_FILES = "Bellows bellows Bellows.py bellows.py".split()
        load_list = []
        
        # Looking for the entry Trawfile.
        if self.opts.bellows is not None:
            BELLOWS_FILES = [self.opts.bellows]
        for bf in BELLOWS_FILES:
            if os.path.isfile(bf):
                return bf

        # Try searching upwards for the main Trawfile
        if not len(load_list) and self.opts.srchup:
            (updir, ignore) = os.path.split(os.getcwd())
            while updir and not len(load_list):
                for bf in BELLOWS_FILES:
                    bf = os.path.join(updir, bf)
                    if os.path.exists(bf):
                        return bf
                (upupdir, ignore) = os.path.split(updir)
                if upupdir == updir: break
                updir = upupdir
        
        raise exc.NoBellowsFileError()
    
    def require(self, filename):
        self.required_files.append(filename)            
    
    def display_tasks(self, patterns):
        names = sorted(self.mgr.tasks.keys())
        tasks = []
        if len(patterns):
            patterns = map(re.compile, patterns)
            for name in names:
                if any(map(lambda p: p.search(name), patterns)):
                    tasks.append(self.mgr.tasks[name])
        else:
            for name in names:
                tasks.append(self.mgr.tasks[name])
        tasks = filter(lambda t: t.descr, tasks)
        if not len(tasks):
            print "No tasks have been described."
            return
        print "Tasks: "
        for t in tasks:
            if not t.descr:
                continue
            descr = (t.descr.strip().splitlines() or [""])[0][:50]
            if len(t.name) < 20:
                print "%-20s # %s" % (t.name, descr)
            else:
                print "%s\n%20s # %s" % (t.name, "", descr)

    def is_dry_run(self):
        return getattr(self.opts, "dryrun", False)

    def rule_depth(self):
        return getattr(self.opts, "rule_depth", 32)

    def log_output(self, task, rval):
        "For debugging task execution in tests."
        self._dbg.append((task, rval))

    def trace(self, mesg):
        # Weird logic to let tests trace with no opts
        if self.opts and not self.opts.trace:
            return
        self.log(mesg)
    
    def log(self, mesg):
        # Weird logic to let tests log with no opts
        if self.opts and not self.opts.verbose:
            return
        # Let nosetests capture output.
        if not self.opts:
            sys.stdout.write("%s\n" % mesg)
        else:
            sys.stderr.write("%s\n" % mesg)
    

application = Application()

def options():
    return [
        op.make_option('-n', dest="dryrun", default=False, action="store_true",
            help="Do a dry run without executing actions."),

        op.make_option('-T', dest="list", default=False, action='store_true',
            metavar="PATTERN",
            help="List the tasks matching PATTERN, then exit."),
        op.make_option('-d', dest="do_deps", default=False, action="store_true",
            help="Display the tasks and dependencies, then exit."),

        op.make_option('-D', dest='rule_depth', default=32, type='int',
            metavar="INTEGER",
            help="Maximum recursion depth for rule resolution."),

        op.make_option('-b', dest="bellows", default=None, metavar="FILE",
            help="Use FILE as the Bellows file."),
        op.make_option('-N', dest="srchup", default=True, action='store_false',
            help="Do not search parent directories for a Bellows file."),
        op.make_option('-G', dest="incsys", default=True, action="store_false",
            help="Ignore system Bellows files."),
        op.make_option('-g', dest="sysglob", default="~/.smithy/*.py",
            metavar="GLOB",
            help="Import system Bellows files. [Default %default]"),
        op.make_option('-L', dest="libglob", default='./smithy/*.py',
            metavar="GLOB",
            help="Import local Bellows files. [Default %default]"),

        op.make_option('-I', dest="incdirs", default=[], action="append",
            metavar="DIR", help="Add DIR to the PYTHONPATH"),

        op.make_option('-t', dest="trace", default=False, action="store_true",
            help="Enable task execution tracing."),
        op.make_option('-v', dest="verbose", default=True, action="store_true",
            help="Display log message on stderr."),
        op.make_option('-q', dest="verbose", action="store_false",
            help="Do not display log message on stderr.")
    ]

def main():
    global application
    parser = op.OptionParser(usage=__usage__, option_list=options())
    opts, args = parser.parse_args()
    try:
        application.run(opts, args)
    except exc.SmithyError, e:
        print str(e)





# Copyright 2010 Paul J. Davis <paul.joseph.davis@gmail.com>
#
# This file is part of the Smithy package released under the MIT license.

import os
import t

@t.test
def test_basic_file():
    fname = t.tmpfname()
    @t.build(fname)
    def build(t):
        open(t.name, "wb").write("o")
    t.app.run(None, [fname])
    t.eq(t.app._dbg, [(fname, None)])
    t.eq(os.path.exists(fname), True)
    t.eq(open(fname).read(), "o")

@t.test
def test_file_as_dep():
    fname = t.tmpfname()
    @t.build(fname)
    def build(t):
        open(t.name, "wb").write("h")
    @t.task([fname])
    def foo():
        return 2
    t.app.run(None, ["foo"])
    t.eq(t.app._dbg, [(fname, None), ("foo", 2)])
    t.eq(os.path.exists(fname), True)
    t.eq(open(fname).read(), "h")

@t.test
def test_file_creation():
    fname = t.tmpfname()
    @t.build(fname, recreate=False)
    def build(task):
        open(task.name, "wb").write(t.uuid.uuid4().hex)
    t.app.run(None, [fname])
    t.eq(t.app._dbg, [(fname, None)])
    t.eq(os.path.exists(fname), True)
    first_run = open(fname).read()
    t.app.clear()
    t.app.run(None, [fname])
    t.eq(t.app._dbg, [])
    t.eq(os.path.exists(fname), True)
    t.eq(open(fname).read(), first_run)

@t.test
def test_not_rebuilt():
    src = t.tmpfname()
    tgt = t.tmpfname()
    @t.build(src, recreate=False)
    def mksrc(task):
        open(task.name, "wb").write(t.uuid.uuid4().hex)
    @t.build(tgt, [src])
    def mktgt(task):
        print task.sources
        print task.source
        open(task.name, "wb").write(open(task.source).read())
    print t.app.mgr.tasks
    t.app.run(None, [tgt])
    t.eq(t.app._dbg, [(src, None), (tgt, None)])
    t.eq(os.path.exists(src), True)
    t.eq(os.path.exists(tgt), True)
    first_run = open(src).read()
    t.eq(open(tgt).read(), first_run)
    t.app.clear()
    t.app.run(None, [tgt])
    t.eq(t.app._dbg, [])
    t.eq(open(src).read(), first_run)
    t.eq(open(tgt).read(), first_run)

@t.test
def test_rerun_build():
    tgt = t.tmpfname()
    @t.build(tgt)
    def mktgt(task):
        task.name.write_bytes("yup")
    
    t.app.run(None, [tgt])
    t.eq(t.app._dbg, [(tgt, None)])
    t.app.clear()
    t.app.run(None, [tgt])
    t.eq(t.app._dbg, [(tgt, None)])

@t.test
def test_dep_not_filetask():
    src = t.tmpfname()
    tgt = t.tmpfname()
    @t.task
    def mksrc(task):
        src.write_bytes("sourcey")
    @t.build(tgt, [mksrc])
    def mktgt(task):
        task.name.write_bytes("yup")
    t.app.run(None, [tgt])
    t.eq(t.app._dbg, [("mksrc", None), (tgt, None)])
    
    t.app.clear()
    t.app.run(None, [tgt])
    t.eq(t.app._dbg, [("mksrc", None), (tgt, None)])

